// [[file:../arroyo-native-parser.org::*The Parser][The Parser:1]]
use anyhow::Result;
use itertools::Itertools;
use lexpr;
use orgize::Event;
use orgize::Org;
use std::boxed::Box;
use std::collections::HashMap;
use std::{error::Error, fs};
// use std::collections::HashMap;

use crate::types::{Document, Heading, InvalidDocError, Keyword, Link};
// The Parser:1 ends here

// [[file:../arroyo-native-parser.org::*The public interface][The public interface:1]]
pub fn parse_document(path: String) -> Result<Document> {
    let org = String::from_utf8(fs::read(path.clone())?).unwrap();
    let org_tree = &Org::parse(&org);

    let keywords = extract_metadata(path.clone(), org_tree)?;
    let headings = extract_headings(path.clone(), org_tree)?;

    Ok(Document {
        path,
        headings,
        keywords,
    })
}
// The public interface:1 ends here

// [[file:../arroyo-native-parser.org::*Extracting Arroyo Keywords][Extracting Arroyo Keywords:1]]
pub fn extract_metadata(path: String, tree: &Org) -> Result<Vec<Keyword>> {
    Ok(tree
        .keywords()
        .map(|kw| Keyword {
            file: path.clone(),
            keyword: kw.key.to_string(),
            value: kw.value.to_string(),
        })
        .collect())
}
// Extracting Arroyo Keywords:1 ends here

// [[file:../arroyo-native-parser.org::*Extracting Arroyo Headings][Extracting Arroyo Headings:1]]
pub fn extract_headings(path: String, tree: &Org) -> Result<Vec<Heading>> {
// Extracting Arroyo Headings:1 ends here

// [[file:../arroyo-native-parser.org::*Extracting Arroyo Headings][Extracting Arroyo Headings:2]]
let mut in_drawer: bool = false;
let mut id_crumbs: Vec<Option<String>> = Vec::new();
let mut cur_id: Option<String> = None;
let mut cur_level: usize = 0;
let mut headings: Vec<Heading> = Vec::new();
headings.push(Heading::default());
let mut links: HashMap<String, Vec<Link>> = HashMap::new();
let mut inherited_tags: Vec<Vec<String>> = Vec::new();
// Extracting Arroyo Headings:2 ends here

// [[file:../arroyo-native-parser.org::*Extracting Arroyo Headings][Extracting Arroyo Headings:3]]
// file level metadata + filetags
let file_metadata = extract_metadata(path.clone(), tree)?;
let filetags = match file_metadata
    .iter()
    .find(|kw| kw.keyword.to_lowercase() == "filetags")
{
    Some(kw) => kw
        .value
        .split(':')
        .map(|s| s.to_string())
        .filter(|s| !s.is_empty())
        .collect(),
    _ => Vec::<String>::new(),
};
headings[0].tags = Some(filetags.clone());
// Extracting Arroyo Headings:3 ends here

// [[file:../arroyo-native-parser.org::*Extracting Arroyo Headings][Extracting Arroyo Headings:4]]
// Extract document title and apply to level 0 heading
let doc_title = match file_metadata
    .iter()
    .find(|kw| kw.keyword.to_lowercase() == "title")
{
    Some(kw) => kw.value.clone(),
    _ => String::from(""),
};
headings[0].text = doc_title;
// Extracting Arroyo Headings:4 ends here

// [[file:../arroyo-native-parser.org::*Extracting Arroyo Headings][Extracting Arroyo Headings:5]]
// state machine go brrr
tree.iter()
    .map(|event| {
        match event {
// Extracting Arroyo Headings:5 ends here

// [[file:../arroyo-native-parser.org::*Heading parser][Heading parser:1]]
Event::Start(orgize::Element::Title(title)) => {
    let properties = title.properties.clone().into_hash_map();
    cur_id = properties.get("ID").map(|id| id.clone().into());

    id_crumbs.truncate(cur_level + 1);
    id_crumbs.push(cur_id.clone());

    let refs = properties
        .get("ROAM_REFS")
        .map(|s| split_quoted_string(s.to_string()).ok())
        .unwrap_or(Some(vec![]));
    let aliases = properties
        .get("ROAM_ALIASES")
        .map(|s| split_quoted_string(s.to_string()).ok())
        .unwrap_or(Some(vec![]));
    cur_level = title.level;

    // reset the tags table
    inherited_tags.truncate(cur_level - 1);
    let new_tags: Vec<String> = title
        .tags
        .iter()
        .map(|mbox| mbox.clone().to_string())
        .collect();
    inherited_tags.push(new_tags);

    let most_tags = inherited_tags.concat();
    let all_tags: Vec<String> = [filetags.clone(), most_tags].concat();

    let h = Heading {
        id: cur_id.clone(),
        level: cur_level,
        text: title.raw.to_string(),
        tags: match all_tags.len() {
            0 => None,
            _ => Some(all_tags),
        },
        refs,
        aliases,
        ..Default::default()
    };
    headings.push(h);
    Ok(())
}
// Heading parser:1 ends here

// [[file:../arroyo-native-parser.org::*File-level Property Drawer parsing][File-level Property Drawer parsing:1]]
Event::Start(orgize::Element::Drawer(drawer)) => {
    in_drawer = drawer.name == "PROPERTIES" && headings[0].id.is_none();
    Ok(())
}
// File-level Property Drawer parsing:1 ends here

// [[file:../arroyo-native-parser.org::*File-level Property Drawer parsing][File-level Property Drawer parsing:2]]
Event::Start(orgize::Element::Text { value }) => {
    if in_drawer {
        // this is where we rely on forked orgize
        let (_, prop_drawer): (_, orgize::elements::PropertiesMap) =
            orgize::elements::Drawer::parse_drawer_content(value)
            .expect("failed to parse properties drawer");
        let properties = prop_drawer.into_hash_map();

        // update cur_id and heading 0 ID since this is
        // implied to be the first drawer, but it's kind
        // of :yikes: to think about it like that! we
        // could be genious enough to have a floating
        // PROPERTIES drawer that would muck things up
        cur_id = properties.get("ID").map(|s| s.to_string());
        if cur_id.is_none() {
            cur_id = properties.get("CUSTOM_ID").map(|s| s.to_string())
        }

        id_crumbs = vec![cur_id.clone()];
        headings[0].id = cur_id.clone();

        headings[0].aliases = properties
            .get("ROAM_ALIASES")
            .map(|s| split_quoted_string(s.to_string()).ok())
            .unwrap_or(Some(vec![]));
        headings[0].refs = properties
            .get("ROAM_REFS")
            .map(|s| split_quoted_string(s.to_string()).ok())
            .unwrap_or(Some(vec![]));
    }

    if headings[0].id.is_none() {
        return Err(InvalidDocError::new_err(
            format!("Root ID is None in {}", path)
        ));
    }

    Ok(())
}
// File-level Property Drawer parsing:2 ends here

// [[file:../arroyo-native-parser.org::*File-level Property Drawer parsing][File-level Property Drawer parsing:3]]
Event::End(orgize::Element::Drawer(_drawer)) => {
    in_drawer = false;
    Ok(())
}
// File-level Property Drawer parsing:3 ends here

// [[file:../arroyo-native-parser.org::*Link parsing][Link parsing:1]]
// Stash links outside the match block in a HashMap shape
// of heading id -> list of links; it would be nice if the
// match block returned an Option<Link> but that doesn't
// play well with the rest of the state machine
Event::Start(orgize::Element::Link(link)) => {
    let dest = link.path.to_string();
    let proto = dest.split_once(':').map(|(proto, _)| proto);

    let last_non_none = match id_crumbs.iter().rev().find_map(|x| x.clone()) {
        Some(last_non_none) => last_non_none,
        None => {
            return Err(InvalidDocError::new_err(
                format!("no non-none ID in {}", path)
            ));
        }
    };

    let link_list = links.entry(last_non_none.clone()).or_insert(Vec::new());
    link_list.push(Link {
        from_file: path.clone().to_string(),
        from_id: last_non_none.clone(),
        to: dest.clone(),
        to_proto: proto.map(String::from),
        text: link.desc.clone().map(String::from),
    });
    Ok(())
}
// Link parsing:1 ends here

// [[file:../arroyo-native-parser.org::*Cleaning up][Cleaning up:1]]
                _ => Ok(()),
            }
        })
        .fold_ok(true, |_, _result| true)?;

    return Ok(headings
        .iter()
        .map(|heading| {
            let mut h = heading.clone();
            h.links = links
                .get(&h.id.clone().unwrap_or(String::from("")))
                .cloned();
            h
        })
        .collect());
}
// Cleaning up:1 ends here

// [[file:../arroyo-native-parser.org::*=split_quoted_string=][=split_quoted_string=:1]]
fn split_quoted_string(quoted_str: String) -> Result<Vec<String>, Box<dyn Error>> {
    let str_as_list = format!("[{}]", &quoted_str);
    let mut sexp =
        lexpr::parse::Parser::from_str_custom(&str_as_list, lexpr::parse::Options::elisp());
    let ret: Vec<String> = sexp
        .value_iter()
        .filter_map(|val| {
            let ret: Option<Vec<String>> = match val {
                Ok(lexpr::Value::Null) => None,
                Ok(lexpr::Value::Vector(values)) => Some(
                    values
                        .iter()
                        .map(|inner_val| match inner_val {
                            lexpr::Value::String(string) => string.to_string(),
                            lexpr::Value::Symbol(string) => string.to_string(),
                            others => todo!("{:?}", others),
                        })
                        .collect(),
                ),
                Ok(lexpr::Value::Symbol(sym)) => Some(vec![sym.to_string()]),
                Err(_) => todo!(),
                val => todo!("{:?}", val),
            };
            ret
        })
        .flatten()
        .collect();

    Ok(ret)
}
// =split_quoted_string=:1 ends here

// [[file:../arroyo-native-parser.org::*Tests][Tests:1]]
#[cfg(test)]
mod tests {
    use std::assert_eq;

    use crate::parse::*;

    #[test]
    fn test_split_quoted_string() {
        assert_eq!(
            split_quoted_string(String::from("")).expect("Could not parse"),
            Vec::<String>::new()
        );

        assert_eq!(
            split_quoted_string(String::from("CCE")).expect("Could not parse"),
            vec!["CCE"]
        );

        assert_eq!(
            split_quoted_string(String::from("\"CCE but with spaces\"")).expect("Could not parse"),
            vec!["CCE but with spaces"]
        );

        assert_eq!(
            split_quoted_string(String::from("\"CCE but with spaces\" \"And 2\""))
                .expect("Could not parse"),
            vec!["CCE but with spaces", "And 2"]
        );

        assert_eq!(
            split_quoted_string(String::from("\"CCE but with spaces\" CCE"))
                .expect("Could not parse"),
            vec!["CCE but with spaces", "CCE"]
        );
    }

    // test for tag inheritence and filetags
    // test for keyword extraction
    // test for link parsing

    // A fairly complicated doc
    #[test]
    fn test_iter_doc_cce() {
        let doc = parse_document(String::from("/home/rrix/org/cce/cce.org"))
            .expect("Did not parse a doc");

        assert_eq!(doc.headings.len(), 4, "Expected four headings in doc");

        assert_eq!(
            doc.headings[0].text, "The Complete Computing Environment",
            "Title correctly set"
        );

        assert_eq!(
            doc.headings[1].text, "[[id:20211219T184243.333209][\"Hey Smell This\"]]",
            "Title correctly set"
        );

        assert!(
            doc.headings[0].links.as_ref().unwrap().len() > 0,
            "No links!"
        );

        assert_eq!(doc.keywords[0].keyword, "TITLE", "Title at da top");
    }
}
// Tests:1 ends here
