// [[file:../arroyo-native-parser.org::*The Parser Document Types][The Parser Document Types:1]]
use pyo3::prelude::*;
use pyo3::pyclass;
use pyo3::exceptions::PyException;

use std::fmt;

pyo3::create_exception!(arroyo_rs, InvalidDocError, PyException);
// The Parser Document Types:1 ends here

// [[file:../arroyo-native-parser.org::*Document][Document:1]]
#[derive(Debug, Clone, Default)]
#[pyclass(dict)]
pub struct Document {
    #[pyo3(get)]
    pub path: String,
    #[pyo3(get)]
    pub headings: Vec<Heading>,
    #[pyo3(get)]
    pub keywords: Vec<Keyword>,
}
// Document:1 ends here

// [[file:../arroyo-native-parser.org::*Document][Document:2]]
impl fmt::Display for Document {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "OrgDoc(from {} with {} headings and {} keywords)",
            self.path,
            self.headings.len(),
            self.keywords.len()
        )
    }
}
// Document:2 ends here

// [[file:../arroyo-native-parser.org::*Document][Document:3]]
#[pymethods]
impl Document {
    pub fn __repr__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Ok(slf.to_string())
    }

    pub fn __str__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Self::__repr__(slf)
    }
}
// Document:3 ends here

// [[file:../arroyo-native-parser.org::*Keyword][Keyword:1]]
#[derive(Debug, Clone)]
#[pyclass(dict)]
pub struct Keyword {
    #[pyo3(get)]
    pub file: String,
    #[pyo3(get)]
    pub keyword: String,
    #[pyo3(get)]
    pub value: String,
}

impl fmt::Display for Keyword {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Keyword(file={}, {}={})",
            self.file, self.keyword, self.value
        )
    }
}

#[pymethods]
impl Keyword {
    pub fn __repr__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Ok(slf.to_string())
    }

    pub fn __str__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Self::__repr__(slf)
    }
}
// Keyword:1 ends here

// [[file:../arroyo-native-parser.org::*Heading][Heading:1]]
#[derive(Debug, Clone, Default)]
#[pyclass(dict)]
pub struct Heading {
    // note that some Headlines may not have an ID, but for purpose of
    // arcology we only care about ones with ID
    #[pyo3(get)]
    pub id: Option<String>,
    #[pyo3(get)]
    pub level: usize,
    #[pyo3(get)]
    pub text: String,
    #[pyo3(get)]
    pub tags: Option<Vec<String>>,
    #[pyo3(get)]
    pub refs: Option<Vec<String>>,
    #[pyo3(get)]
    pub aliases: Option<Vec<String>>,

    #[pyo3(get)]
    pub links: Option<Vec<Link>>,
}

impl fmt::Display for Heading {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Heading(id={}, title={}, {} tags, {} refs, {} aliases, {} links)",
            self.id.clone().unwrap_or("None".to_owned()),
            self.text,
            self.tags.clone().unwrap_or(vec![]).len(),
            self.refs.clone().unwrap_or(vec![]).len(),
            self.aliases.clone().unwrap_or(vec![]).len(),
            self.links.clone().unwrap_or(vec![]).len(),
        )
    }
}

#[pymethods]
impl Heading {
    pub fn __repr__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Ok(slf.to_string())
    }

    pub fn __str__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Self::__repr__(slf)
    }
}
// Heading:1 ends here

// [[file:../arroyo-native-parser.org::*Link][Link:1]]
#[derive(Debug, Clone, Default)]
#[pyclass(dict)]
pub struct Link {
    #[pyo3(get)]
    pub from_file: String,
    #[pyo3(get)]
    pub from_id: String,
    #[pyo3(get)]
    pub to: String,
    #[pyo3(get)]
    pub to_proto: Option<String>,
    #[pyo3(get)]
    pub text: Option<String>,
}

impl fmt::Display for Link {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Link({}#{} -> [[{}][{}]])",
            self.from_file,
            self.from_id.clone(),
            self.to,
            self.text.clone().unwrap_or("".to_owned())
        )
    }
}

#[pymethods]
impl Link {
    pub fn __repr__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Ok(slf.to_string())
    }

    pub fn __str__(slf: PyRef<'_, Self>) -> PyResult<String> {
        Self::__repr__(slf)
    }
}
// Link:1 ends here
